// ImgWnd.cpp : t@C
//

#include "stdafx.h"
#include "ClipMM.h"
#include "ImgWnd.h"


// CImgWnd

IMPLEMENT_DYNAMIC(CImgWnd, CWnd)

CImgWnd::CImgWnd()
	: m_pImage(NULL)
	, m_xScroll(0)
	, m_yScroll(0)
	, m_cxImage(0)
	, m_cyImage(0)
	, m_bCapture(FALSE)
{

}

CImgWnd::~CImgWnd()
{
}

BOOL CImgWnd::Create(DWORD dwStyle0, CWnd* pWndParent, UINT nID)
{
	DWORD dwStyle = dwStyle0 | WS_HSCROLL | WS_VSCROLL;

	BOOL b = 
	CWnd::Create(AfxRegisterWndClass(CS_DBLCLKS, AfxGetApp()->LoadStandardCursor(IDC_ARROW)), 
		L"", 
		dwStyle,
		CRect(0, 0, 0, 0),
		pWndParent,
		nID,
		NULL);

	return b;
}

void CImgWnd::SetImage(Gdiplus::Image* pImage)
{
	m_pImage = pImage;
	m_xScroll = m_yScroll = 0;

	if (m_hWnd)
	{
		CalcScrollParams();
		Invalidate();
	}
}

BEGIN_MESSAGE_MAP(CImgWnd, CWnd)
	ON_WM_CREATE()
	ON_WM_DESTROY()
	ON_WM_SIZE()
	ON_WM_PAINT()
	ON_WM_HSCROLL()
	ON_WM_VSCROLL()
	ON_WM_LBUTTONDOWN()
	ON_WM_LBUTTONUP()
	ON_WM_RBUTTONDOWN()
	ON_WM_MOUSEMOVE()
	ON_WM_MOUSEWHEEL()
	ON_WM_KEYDOWN()
	ON_WM_GETDLGCODE()
END_MESSAGE_MAP()



// CImgWnd bZ[W nh[




int CImgWnd::OnCreate(LPCREATESTRUCT lpCreateStruct)
{
	if (CWnd::OnCreate(lpCreateStruct) == -1)
		return -1;

	m_cxMargin = m_cyMargin = 5;
	m_cxLine = m_cyLine = 8;

	CalcScrollParams();


	return 0;
}

void CImgWnd::CalcScrollParams()
{
	CRect rcClient;
	GetClientRect(rcClient);

	m_cxClient = rcClient.Width();
	m_cyClient = rcClient.Height();

	if (m_pImage)
	{
		m_cxImage = m_pImage->GetWidth();
		m_cyImage = m_pImage->GetHeight();
	}
	else
	{
		m_cxImage = 0;
		m_cyImage = 0;
	}

	int cxTotal = m_cxImage + m_cxMargin * 2;
	m_xScrollMax = max(cxTotal - m_cxClient, 0);

	if (m_xScroll > m_xScrollMax)
		m_xScroll = m_xScrollMax;

	SCROLLINFO si = {0};
	si.cbSize = sizeof(si);
	si.fMask = SIF_PAGE | SIF_POS | SIF_RANGE;
	si.nMin = 0;
	si.nMax = cxTotal - 1;
	si.nPos = m_xScroll;
	si.nPage = m_cxClient;
	SetScrollInfo(SB_HORZ, &si);

	int cyTotal = m_cyImage + m_cyMargin * 2;
	m_yScrollMax = max(cyTotal - m_cyClient, 0);

	if (m_yScroll > m_yScrollMax)
		m_yScroll = m_yScrollMax;

	si.nMin = 0;
	si.nMax = cyTotal - 1;
	si.nPos = m_yScroll;
	si.nPage = m_cyClient;
	SetScrollInfo(SB_VERT, &si);
}

void CImgWnd::OnDestroy()
{
	CWnd::OnDestroy();
}


void CImgWnd::OnSize(UINT nType, int cx, int cy)
{
	CWnd::OnSize(nType, cx, cy);

	CalcScrollParams();
	Invalidate();
}


void CImgWnd::OnPaint()
{
	CPaintDC dc(this); // device context for painting

	CRect rcClip;
	dc.GetClipBox(rcClip);

	Gdiplus::Graphics gWindow(dc.GetSafeHdc());
	Gdiplus::SolidBrush br(Gdiplus::Color(238, 242, 250));

	Gdiplus::Bitmap bmPlane(rcClip.Width(), rcClip.Height(), PixelFormat32bppRGB);
	Gdiplus::Graphics g(&bmPlane);
	Gdiplus::PointF ptClip((Gdiplus::REAL)rcClip.left, (Gdiplus::REAL)rcClip.top);
	g.TranslateTransform(-ptClip.X, -ptClip.Y);

	g.FillRectangle(&br, rcClip.left, rcClip.top, rcClip.Width(), rcClip.Height());
	if (m_pImage)
	{
		CRect rcImage;

		rcImage.left = m_cxMargin - m_xScroll;
		rcImage.top = m_cyMargin - m_yScroll;
		rcImage.right = rcImage.left + m_cxImage;
		rcImage.bottom = rcImage.top + m_cyImage;

		CRect rc;
		if (rc.IntersectRect(rcClip, rcImage))
		{
			Gdiplus::PointF pt;
			pt.X = (Gdiplus::REAL)rcImage.left;
			pt.Y = (Gdiplus::REAL)rcImage.top;
			g.DrawImage(m_pImage, rcImage.left, rcImage.top, m_cxImage, m_cyImage);
		}
	}
	gWindow.DrawImage(&bmPlane, ptClip);
}


void CImgWnd::OnHScroll(UINT nSBCode, UINT nPos, CScrollBar* pScrollBar)
{
	int xNew = m_xScroll;
	BOOL bSetPos = TRUE;
	BOOL bForceSetPos = FALSE;

	switch (nSBCode)
	{
	case SB_ENDSCROLL:		// Ends scroll.
		break;
	case SB_LEFT:			// Scrolls to the upper left.
		xNew = 0;
		break;
	case SB_RIGHT:			// Scrolls to the lower right.
		xNew = m_xScrollMax;
		break;
	case SB_LINELEFT:		// Scrolls left by one unit.
		xNew = max(xNew - m_cxLine, 0);
		break;
	case SB_LINERIGHT:		// Scrolls right by one unit.
		xNew = min(xNew + m_cxLine, m_xScrollMax);
		break;
	case SB_PAGELEFT:		// Scrolls left by the width of the window.
		xNew = max(xNew - m_cxClient, 0);
		break;
	case SB_PAGERIGHT:		// Scrolls right by the width of the window.
		xNew = min(xNew + m_cxClient, m_xScrollMax);
		break;
	case SB_THUMBPOSITION:	// The user has dragged the scroll box (thumb) and released the mouse button. The HIWORD indicates the position of the scroll box at the end of the drag operation.
	case SB_THUMBTRACK:		// The user is dragging the scroll box. This message is sent repeatedly until the user releases the mouse button. The HIWORD indicates the position that the scroll box has been dragged to.
		{
			SCROLLINFO si = {0};
			si.cbSize = sizeof(si);
			si.fMask = SIF_TRACKPOS;
			GetScrollInfo(SB_HORZ, &si);
			xNew = si.nTrackPos;
			if (nSBCode == SB_THUMBTRACK)
				bSetPos = FALSE;
			else
				bForceSetPos = TRUE;
		}
	}

	if (xNew != m_xScroll || bForceSetPos)
	{
		Scroll(xNew, m_yScroll, bSetPos, FALSE);
		UpdateWindow();
	}

	CWnd::OnHScroll(nSBCode, nPos, pScrollBar);
}


void CImgWnd::OnVScroll(UINT nSBCode, UINT nPos, CScrollBar* pScrollBar)
{
	int yNew = m_yScroll;
	BOOL bSetPos = TRUE;
	BOOL bForceSetPos = FALSE;

	CRect rcClient;
	GetClientRect(rcClient);


	switch (nSBCode)
	{
	case SB_BOTTOM:		// Scrolls to the lower right.
		yNew = m_yScrollMax;
		break;
	case SB_ENDSCROLL:	// Ends scroll.
		break;
	case SB_LINEDOWN:	// Scrolls one line down.
		yNew = min(yNew + m_cyLine, m_yScrollMax);
		break;
	case SB_LINEUP:		// Scrolls one line up.
		yNew = max(yNew - m_cyLine, 0);
		break;
	case SB_PAGEDOWN:	// Scrolls one page down.
		yNew = min(yNew + rcClient.Height(), m_yScrollMax);
		break;
	case SB_PAGEUP:		// Scrolls one page up.
		yNew = max(yNew - rcClient.Height(), m_yScrollMax);
		break;
	case SB_THUMBPOSITION:	// The user has dragged the scroll box (thumb) and released the mouse button. The HIWORD indicates the position of the scroll box at the end of the drag operation.
	case SB_THUMBTRACK:	// The user is dragging the scroll box. This message is sent repeatedly until the user releases the mouse button. The HIWORD indicates the position that the scroll box has been dragged to.
		{
			SCROLLINFO si = {0};
			si.cbSize = sizeof(si);
			si.fMask = SIF_TRACKPOS;
			GetScrollInfo(SB_VERT, &si);
			yNew = si.nTrackPos;
			if (nSBCode == SB_THUMBTRACK)
				bSetPos = FALSE;
			else
				bForceSetPos = TRUE;
		}
		break;
	case SB_TOP:
		yNew = 0;
		break;
	}

	if (yNew != m_yScroll || bForceSetPos)
	{
		Scroll(m_xScroll, yNew, FALSE, bSetPos);
		UpdateWindow();
	}
	CWnd::OnVScroll(nSBCode, nPos, pScrollBar);
}

void CImgWnd::Scroll(int xScrollNew, int yScrollNew, BOOL bSetHBar, BOOL bSetVBar)
{
	int cxScroll = m_xScroll - xScrollNew;
	int cyScroll = m_yScroll - yScrollNew;

	m_xScroll = xScrollNew;
	m_yScroll = yScrollNew;

	ScrollWindow(cxScroll, cyScroll);

	if (bSetHBar)
	{
		SCROLLINFO si = {0};
		si.cbSize = sizeof(si);
		si.fMask = SIF_POS;
		si.nPos = xScrollNew;
		SetScrollInfo(SB_HORZ, &si, TRUE);
	}

	if (bSetVBar)
	{
		SCROLLINFO si = {0};
		si.cbSize = sizeof(si);
		si.fMask = SIF_POS;
		si.nPos = yScrollNew;
		SetScrollInfo(SB_VERT, &si, TRUE);
	}
}


void CImgWnd::OnLButtonDown(UINT nFlags, CPoint point)
{
	m_bCapture = TRUE;
	m_ptMouse = point;
	m_xScrollMouse = m_xScroll;
	m_yScrollMouse = m_yScroll;
	
	SetCapture();


	CWnd::OnLButtonDown(nFlags, point);
}


void CImgWnd::OnLButtonUp(UINT nFlags, CPoint point)
{
	m_bCapture = FALSE;
	ReleaseCapture();

	CWnd::OnLButtonUp(nFlags, point);
}


void CImgWnd::OnRButtonDown(UINT nFlags, CPoint point)
{
	CWnd::OnRButtonDown(nFlags, point);
}


void CImgWnd::OnMouseMove(UINT nFlags, CPoint point)
{
	if (m_bCapture)
	{
		CPoint ptRel = point - m_ptMouse;

		int xScrollNew = min(max(m_xScrollMouse - ptRel.x, 0), m_xScrollMax);
		int yScrollNew = min(max(m_yScrollMouse - ptRel.y, 0), m_yScrollMax);

		Scroll(xScrollNew, yScrollNew, TRUE, TRUE);
		UpdateWindow();
	}

	CWnd::OnMouseMove(nFlags, point);
}


BOOL CImgWnd::OnMouseWheel(UINT nFlags, short zDelta, CPoint pt)
{
	int yRel = -zDelta / 5;
	int yScrollNew = min(max(m_yScroll + yRel, 0), m_yScrollMax);

	Scroll(m_xScroll, yScrollNew, FALSE, TRUE);

	return CWnd::OnMouseWheel(nFlags, zDelta, pt);
}


void CImgWnd::OnKeyDown(UINT nChar, UINT nRepCnt, UINT nFlags)
{
	int xScrollNew = m_xScroll;
	int yScrollNew = m_yScroll;

	switch (nChar)
	{
	case VK_UP:
		yScrollNew = max(yScrollNew - m_cyLine, 0);
		break;
	case VK_DOWN:
		yScrollNew = min(yScrollNew + m_cyLine, m_yScrollMax);
		break;
	case VK_LEFT:
		xScrollNew = max(xScrollNew - m_cxLine, 0);
		break;
	case VK_RIGHT:
		xScrollNew = min(xScrollNew + m_cxLine, m_xScrollMax);
		break;
	case VK_PRIOR:
		yScrollNew = max(yScrollNew - m_cyClient, 0);
		break;
	case VK_NEXT:
		yScrollNew = min(yScrollNew + m_cyClient, m_yScrollMax);
		break;
	case VK_HOME:
		if (GetKeyState(VK_CONTROL) < 0)
			yScrollNew = 0;
		xScrollNew = 0;
		break;
	case VK_END:
		if (GetKeyState(VK_CONTROL) < 0)
			yScrollNew = m_yScrollMax;
		xScrollNew = m_xScrollMax;
		break;
	}

	Scroll(xScrollNew, yScrollNew, TRUE, TRUE);
	UpdateWindow();

	CWnd::OnKeyDown(nChar, nRepCnt, nFlags);
}


UINT CImgWnd::OnGetDlgCode()
{
//	return CWnd::OnGetDlgCode();
	return DLGC_WANTARROWS;
}
