// TextViewDialog.cpp : t@C
//

#include "stdafx.h"
#include "ClipMM.h"
#include "TextViewDialog.h"
#include "afxdialogex.h"
#include "Util.h"
#include "AddCodePageDialog.h"
#include "GuessCharSet.h"

#pragma comment(linker, "/DEFAULTLIB:UxTheme.lib")

static const BYTE UTF16LEBOM[] = {0xFF, 0xFE};
static const BYTE UTF16BEBOM[] = {0xFE, 0xFF};
static const BYTE UTF8BOM[] = {0xEF, 0xBB, 0xBF};

HMODULE hModuleMLang = NULL;
HRESULT (STDAPICALLTYPE *pfConvertINetMultiByteToUnicode)(LPDWORD lpdwMode, DWORD dwEncoding, LPCSTR lpSrcStr, LPINT lpnMultiCharCount, __out_ecount_opt(*lpnWideCharCount) LPWSTR lpDstStr, __inout_opt LPINT lpnWideCharCount); 

// CTextViewDialog _CAO

IMPLEMENT_DYNAMIC(CTextViewDialog, CClipMMPropertyPage)

CTextViewDialog::CTextViewDialog(CCBData* pData)
	: CClipMMPropertyPage(pData, CTextViewDialog::IDD)
	, m_pbrEditBack(NULL)
	, m_nCodePage(CP_ACP)
{
	hModuleMLang = LoadLibrary(L"mlang.dll");
	if (hModuleMLang)
	{
		*(FARPROC*)&pfConvertINetMultiByteToUnicode = GetProcAddress(hModuleMLang, "ConvertINetMultiByteToUnicode");
	}
}

CTextViewDialog::~CTextViewDialog()
{
	if (m_pbrEditBack)
		delete m_pbrEditBack;

	ResetCPTable();

	if (hModuleMLang)
		FreeLibrary(hModuleMLang);
		
}

void CTextViewDialog::ResetCPTable()
{
	for (int i = 0; i < m_cptable.GetSize(); i++)
	{
		delete m_cptable[i];
	}
	m_cptable.RemoveAll();
}

void CTextViewDialog::MakeCPTable(int nCPCurrentOld)
{
	CClipMMApp* pApp = (CClipMMApp*)AfxGetApp();

	ResetCPTable();
	
	m_cptable.Add(new CMMCP(L"CP_ACP", CP_ACP));
	m_cptable.Add(new CMMCP(L"CP_OEMCP", CP_OEMCP));
	m_cptable.Add(new CMMCP(L"CP_UTF8",	CP_UTF8));
	m_cptable.Add(new CMMCP(L"UTF-16LE", CP_CBDATA_UTF16));
	m_cptable.Add(new CMMCP(L"UTF-16BE", CP_CBDATA_UTF16BE));

	// [Uݒ}[W
	CArray<CMMCP*, CMMCP*> t;
	CString str = pApp->GetProfileString(L"Settings", L"CodePages", g_pszDefaultCodePages);

	CStringArray arrayLines;
	StringArrayFromString(arrayLines, str, L";");
	for (int i = 0; i < arrayLines.GetSize(); i++)
	{
		CString strLine = arrayLines[i];
		CString strName;
		CString strValue;
		int n = strLine.Find(L'=');
		if (n < 0)
		{
			strName = strValue = strLine.Trim();
		}
		else
		{
			strName = strLine.Left(n).Trim();
			strValue = strLine.Mid(n+1).Trim();
		}
		if (strName.IsEmpty() || strValue.IsEmpty())
		{
		}
		else
		{
			CMMCP* p = new CMMCP(strName, _wtoi(strValue));

			// lƂԂĂu
			for (int n = (int)m_cptable.GetSize() - 1; n >= 0; n--)
			{
				if (m_cptable[n]->m_nCodePage == p->m_nCodePage)
				{
					delete m_cptable[n];
					m_cptable.RemoveAt(n);
					break;
				}
			}
			t.Add(p);
		}
	}
	m_cptable.Append(t);

	// m_pDatãftHgĂȂΒǉ
	if (m_pData)
	{
		int n = m_pData->GetSuitableCodePage();
		if (n != CP_CBDATA_UNKNOWN)
		{
			AddCPTable(n);
		}
	}

	// Ƃ̒lĂȂΒǉ
	AddCPTable(nCPCurrentOld);
}

void CTextViewDialog::AddCPTable(int n)
{
	BOOL bFound = FALSE;
	for (int i = 0; i < m_cptable.GetSize(); i++)
	{
		if (m_cptable[i]->m_nCodePage == n)
		{
			bFound = TRUE;
			break;
		}
	}
	if (!bFound)
	{
		CString str;
		str.Format(L"%d", n);
		m_cptable.Add(new CMMCP(str, n));
	}
}

void CTextViewDialog::DoDataExchange(CDataExchange* pDX)
{
	CClipMMPropertyPage::DoDataExchange(pDX);
	DDX_Control(pDX, IDC_COMBO_CODE_PAGE, m_comboCodePage);
	DDX_Control(pDX, IDC_EDIT_TEXT, m_editText);
	DDX_Control(pDX, IDC_STATIC_NUL, m_staticNUL);
}


BEGIN_MESSAGE_MAP(CTextViewDialog, CClipMMPropertyPage)
	ON_WM_SIZE()
	ON_CBN_SELCHANGE(IDC_COMBO_CODE_PAGE, &CTextViewDialog::OnSelchangeComboCodePage)
	ON_CBN_SELENDOK(IDC_COMBO_CODE_PAGE, &CTextViewDialog::OnSelendokComboCodePage)
	ON_BN_CLICKED(IDC_BUTTON_SET, &CTextViewDialog::OnBnClickedButtonSet)
	ON_WM_CTLCOLOR()
	ON_WM_SYSCOLORCHANGE()
END_MESSAGE_MAP()


// CTextViewDialog bZ[W nh[


void CTextViewDialog::OnSize(UINT nType, int cx, int cy)
{
	CClipMMPropertyPage::OnSize(nType, cx, cy);

	if (m_editText.m_hWnd)
	{
		CRect rc;
		m_editText.GetWindowRect(rc);
		ScreenToClient(rc);
		rc.right = cx;
		rc.bottom = cy;

		m_editText.MoveWindow(rc);
	}

}


BOOL CTextViewDialog::OnInitDialog()
{
	CClipMMPropertyPage::OnInitDialog();

//	EnableThemeDialogTexture(m_hWnd, ETDT_DISABLE);

	m_pbrEditBack = new CBrush;
	m_pbrEditBack->CreateSolidBrush(GetSysColor(COLOR_WINDOW));

	BYTE* pMem = NULL;

	if (m_pData)
	{
		if (m_pData->m_cbMem > INT_MAX)
		{
			pMem = NULL;
		}
		else
		{
			pMem = m_pData->LoadMemData();
		}
		if (pMem)
		{
			m_nCodePage = m_pData->GetSuitableCodePage();
			if (m_nCodePage == CP_CBDATA_UNKNOWN)
			{
				int cbBOM;
				int n = CheckBOM(pMem, cbBOM);
				if (n == CP_UTF8)
				{
					m_nCodePage = CP_UTF8;
				}
				else if (n == CP_CBDATA_UTF16)
				{
					m_nCodePage = CP_CBDATA_UTF16;
				}
				else if (n == CP_CBDATA_UTF16BE)
				{
					m_nCodePage = CP_CBDATA_UTF16BE;
				}
				else
				{
					// ANSIUTF16𐄑
					// ԒnNUL̂ƂɃf[^Ă悤ȂANSỈ\͒ႢH
					BOOL bOddNulFound = FALSE;
					BOOL bMaybeUnicode = FALSE;
					for (size_t i = 0; i < min(m_pData->m_cbMem, 16384); i++)
					{
						if (pMem[i])
						{
							// [łȂ
							if (bOddNulFound)
							{
								bMaybeUnicode = TRUE;
								break;
							}
						}
						else
						{
							if ((i % 2) == 1)
								bOddNulFound = TRUE;
						}
					}
					if (bMaybeUnicode)
					{
						m_nCodePage = CP_CBDATA_UTF16;
					}
					else
					{
						size_t nErrorsUTF8 = CountUTF8Errors((const char*)pMem, min((size_t)m_pData->m_cbMem, 16384));
						size_t nErrorsSJIS = CountSJISErrors((const char*)pMem, min((size_t)m_pData->m_cbMem, 16384));

						if (nErrorsUTF8 == 0 && nErrorsSJIS == 0)
						{
							if (GetSJISDoubleByteCharError((const char*)pMem, min((int)m_pData->m_cbMem, 16384)))
							{
								nErrorsSJIS++;		// G[ɂĂ
							}
						}

						if (nErrorsUTF8 && nErrorsSJIS)
							m_nCodePage = CP_CBDATA_UTF16;		// UTF16?
						else if (nErrorsSJIS > nErrorsUTF8)
							m_nCodePage = CP_UTF8;
						else
							m_nCodePage = CP_ACP;
					}
				}
			}
		}
	}

	FillCodePageComboBox();

	ConvertAndSetText(pMem);

	return TRUE;  // return TRUE unless you set the focus to a control
	// O : OCX vpeB y[W͕K FALSE Ԃ܂B
}

int CTextViewDialog::CheckBOM(BYTE* pMem, int& cbBOM)
{
	cbBOM = 0;
	if (m_pData->m_cbMem >= _countof(UTF8BOM) &&
		memcmp(pMem, UTF8BOM, _countof(UTF8BOM)) == 0)
	{
		cbBOM = 3;
		return CP_UTF8;
	}
	else if (m_pData->m_cbMem >= _countof(UTF16LEBOM) &&
		memcmp(pMem, UTF16LEBOM, _countof(UTF16LEBOM)) == 0)
	{
		cbBOM = 2;
		return CP_CBDATA_UTF16;
	}
	else if (m_pData->m_cbMem >= _countof(UTF16BEBOM) &&
		memcmp(pMem, UTF16BEBOM, _countof(UTF16BEBOM)) == 0)
	{
		cbBOM = 2;
		return CP_CBDATA_UTF16BE;
	}
	return CP_CBDATA_UNKNOWN;

}

void CTextViewDialog::FillCodePageComboBox()
{
	CClipMMApp* pApp = (CClipMMApp*)AfxGetApp();

	int nCPCurrentOld = CP_ACP;
	// Jg̑I
	int n = m_comboCodePage.GetCurSel();
	if (n >= 0 && n < m_cptable.GetSize())
	{
		nCPCurrentOld = m_cptable[n]->m_nCodePage;
	}
	else
	{
		if (m_pData)
			nCPCurrentOld = m_nCodePage;
	}

	m_comboCodePage.ResetContent();

	MakeCPTable(nCPCurrentOld);

	for (int i = 0; i < m_cptable.GetSize(); i++)
	{
		m_comboCodePage.AddString(m_cptable[i]->m_strName);
		if (m_cptable[i]->m_nCodePage == nCPCurrentOld)
		{
			m_comboCodePage.SetCurSel(i);
		}
	}
}

/*
f[^[hς݂ȂpMemœnBJB
[hȂNULLnΒŃ[h
*/

void CTextViewDialog::ConvertAndSetText(BYTE* pMem)
{
	if (!pMem)
	{
		if (m_pData)
		{
			if (m_pData->m_cbMem > INT_MAX)
			{
				pMem = NULL;
			}
			else
			{
				pMem = m_pData->LoadMemData();
			}
		}
	}

	CString str;
	BOOL bNUL = FALSE;
	if (pMem)
	{
		int n = m_comboCodePage.GetCurSel();
		if (n >= 0 && n < m_cptable.GetSize())
		{
			m_nCodePage = m_cptable[n]->m_nCodePage;
			if (m_nCodePage >= 0)
			{
				int cbIn = (int)m_pData->m_cbMem;
				// NUL ŏI[Ă邩`FbN
				for (int i = 0; i < m_pData->m_cbMem; i++)
				{
					if (pMem[i] == 0)
					{
						cbIn = i;
						bNUL = TRUE;
						break;
					}
				}
				int cbBOM = 0;
				if (m_nCodePage == CP_UTF8 &&
					CheckBOM(pMem, cbBOM) == CP_UTF8)
				{
				}
				else
				{
					cbBOM = 0;
				}

				if (cbIn - cbBOM >= 0)
				{
					if (m_nCodePage != CP_ACP &&
						m_nCodePage != CP_OEMCP &&
						pfConvertINetMultiByteToUnicode)
					{
						// ̂قgR[hy[WH@01̓_݂B
						DWORD dwMode = 0;
						int cbSrc = cbIn - cbBOM;
						int cchDst = cbSrc + 1024;	// 莚͂Ȃ܂BwȃR[hy[WƂȂH
						WCHAR* pchOutBuf = str.GetBuffer(cchDst);	
						if (SUCCEEDED((*pfConvertINetMultiByteToUnicode)(&dwMode, m_nCodePage, (LPCSTR)(pMem + cbBOM), &cbSrc, pchOutBuf, &cchDst)))
						{
							str.ReleaseBuffer(cchDst);
						}
						else
						{
							str.ReleaseBuffer(0);
						}
					}
					else
					{
						int cch = MultiByteToWideChar(m_nCodePage, 0, (LPCSTR)(pMem + cbBOM), cbIn - cbBOM, NULL, 0);
						WCHAR* pchOutBuf = str.GetBuffer(cch);
						if (pchOutBuf)
						{
							MultiByteToWideChar(m_nCodePage, 0, (LPCSTR)(pMem + cbBOM), cbIn - cbBOM, pchOutBuf, cch);
							str.ReleaseBuffer(cch);
						}
					}
				}
			}
			else
			{
				int cchIn = (int)(m_pData->m_cbMem / 2);
				// NULŏI[Ă邩`FbN
				for (int i = 0; i < m_pData->m_cbMem / 2; i++)
				{
					if (((WCHAR*)pMem)[i] == 0)
					{
						cchIn = i;
						bNUL = TRUE;
						break;
					}
				}
				int cbBOM = 0;
				if (CheckBOM(pMem, cbBOM) == m_nCodePage)
				{
				}
				else
				{
					cbBOM = 0;
				}
				if (m_nCodePage == CP_CBDATA_UTF16BE)
				{
					for (int i = 0; i < m_pData->m_cbMem / 2; i++)
					{
						BYTE b = pMem[i * 2];
						pMem[i * 2] = pMem[i * 2 + 1];
						pMem[i * 2 + 1] = b;
					}
				}

				str = CString((const WCHAR*)(pMem + cbBOM), (int)(cchIn - cbBOM / 2));
			}
		}
		delete[] pMem;
	}
	// CR+LFɂȂĂ邩`FbN
	int nNoLF = 0;
	int nNoCR = 0;
	int nCRLF = 0;
	{
		const WCHAR* pch = (const WCHAR*)str;
		while (*pch)
		{
			if (*pch == L'\r')
			{
				if (*(pch+1) == L'\n')
				{
					nCRLF++;
					pch++;
				}
				else
				{
					nNoLF++;
				}
			}
			else if (*pch == L'\n')
			{
				nNoCR++;
			}
			pch++;
		}
	}
	if (nNoCR > 0 || nNoLF > 0)
	{
		// sR[h␳
		int cchSrc = str.GetLength();
		CString str2;
		WCHAR* pchDst = str2.GetBuffer(cchSrc + nNoCR + nNoLF);
#ifdef _DEBUG
		WCHAR* pchDstStart = pchDst;
#endif
		const WCHAR* pch = (const WCHAR*)str;
		while (*pch)
		{
			if (*pch == L'\r')
			{
				if (*(pch+1) == L'\n')
				{
					*(pchDst++) = *(pch++);
					*(pchDst++) = *(pch++);
				}
				else
				{
					*(pchDst++) = *(pch++);
					*(pchDst++) = L'\n';
				}
			}
			else if (*pch == L'\n')
			{
				*(pchDst++) = L'\r';
				*(pchDst++) = *(pch++);
			}
			else
			{
				*(pchDst++) = *(pch++);
			}
		}
		ASSERT((int)(pchDst - pchDstStart) == cchSrc + nNoCR + nNoLF);
		str2.ReleaseBuffer(cchSrc + nNoCR + nNoLF);
		str = str2;
	}


	m_editText.SetWindowText(str);

	CString strNLType;
	if (nCRLF == 0 && nNoCR == 0 && nNoLF == 0)
		strNLType = L"Ȃ";
	else if (nCRLF > 0 && nNoCR == 0 && nNoLF == 0)
		strNLType = L"CR+LF";
	else if (nCRLF == 0 && nNoCR > 0 && nNoLF == 0)
		strNLType = L"*LF̂*";
	else if (nCRLF == 0 && nNoCR == 0 && nNoLF > 0)
		strNLType = L"*CR̂*";
	else
		strNLType = L"**";
	m_staticNUL.SetWindowText(CString(L"I[NUL: ") + (bNUL ? L"" : L"Ȃ") +
								L"  sR[h: " + strNLType);
}



void CTextViewDialog::OnSelchangeComboCodePage()
{
	ConvertAndSetText(NULL);
}


void CTextViewDialog::OnOK()
{
//	CClipMMPropertyPage::OnOK();
}


void CTextViewDialog::OnSelendokComboCodePage()
{
}

void CTextViewDialog::OnBnClickedButtonSet()
{
	CAddCodePageDialog dlg;
	if (dlg.DoModal() != IDOK)
		return;

	FillCodePageComboBox();
	ConvertAndSetText(NULL);
}


HBRUSH CTextViewDialog::OnCtlColor(CDC* pDC, CWnd* pWnd, UINT nCtlColor)
{
	HBRUSH hbr = CClipMMPropertyPage::OnCtlColor(pDC, pWnd, nCtlColor);

	if (pWnd->GetDlgCtrlID() == IDC_EDIT_TEXT)
	{
		HBRUSH h = (HBRUSH)m_pbrEditBack->GetSafeHandle();
		if (h)
			hbr = h;
	}

	return hbr;
}


void CTextViewDialog::OnSysColorChange()
{
	CClipMMPropertyPage::OnSysColorChange();

	if (m_pbrEditBack)
		delete m_pbrEditBack;

	m_pbrEditBack = new CBrush;
	m_pbrEditBack->CreateSolidBrush(GetSysColor(COLOR_WINDOW));


}


void CTextViewDialog::OnToolSetDefaultView()
{
	if (m_pData)
	{
		m_pData->SetSuitableType(CCBData::suitabletypeText);
	}
}
